from .lib.attrdict import AttrDict

class HTTPResponse:	
	def __init__ (self, response):
		self.response = response
		self.request = response.request
	
	def __del__ (self):
		self.response.request = None
		self.request = None		
		self.response = None		
			
	def __getattr__ (self, name):
		return getattr (self.response, name)



responses = {
		100: "Continue",
		101: "Switching Protocols",
		200: "OK",
		201: "Created",
		202: "Accepted",
		203: "Non-Authoritative Information",
		204: "Deleted",
		205: "Reset Content",
		206: "Partial Content",
		300: "Multiple Choices",
		301: "Moved Permanently",
		302: "Moved Temporarily",
		303: "See Other",
		304: "Not Modified",
		305: "Use Proxy",
		306: "Switch Proxy",
		307: "Temporary Redirect",
		308: "Permanent Redirect",
		400: "Bad Request",
		401: "Unauthorized",
		402: "Payment Required",
		403: "Forbidden",
		404: "Not Found",
		405: "Method Not Allowed",
		406: "Not Acceptable",
		407: "Proxy Authentication Required",
		408: "Request Timeout",
		409: "Conflict",
		410: "Gone",
		411: "Length Required",			
		412: "Precondition Failed",
		413: "Request Entity Too Large",
		414: "Request-URI Too Large",
		415: "Unsupported Media Type",
		500: "Internal Server Error",
		501: "Not Implemented",
		502: "Bad Gateway",
		503: "Service Unavailable",
		504: "Gateway Time-out",
		505: "HTTP Version Not Supported",
		506: "Proxy Error",
		507: "Failed Establishing Connection",
		508: "Application Error",
		700: "Socket Error",
		701: "Exception Occured",
		702: "Socket Timeout",
		703: "Socket Panic",
		704: "DNS Not Found",
		705: "Entered Shutdown Process",
		706: "Unknown Authedentification Method",
		707: "HTTP Header/Body Error",
		708: "No Data Recieved",
		709: "Invalid Content",
		710: "Channel Closed",
		711: "Redirect Error",
		712: "Controlled Shutdown",
		713: "SSL Handshake Failure",
		714: "Connection Failure",
		715: "Response Header Error",
		716: "Protocol Switching Failure",
		717: "Invalid Cotnent Type",
		718: "Unacceptable Content Type",	
		719: "Too Large Content",	
		720: "HTTP2 Connection Closed",
		721: "HTTP2 Stream Reset",
		722: "Object timeout"	
}

def get (code, default = "Undefined Error"):
	return responses.get (code, default)
	
	



from . import db3
import psycopg2
import warnings

warnings.simplefilter('default')
warnings.warn (
   "pg2 will be deprecated, use sqlphile.pg2",
    DeprecationWarning
)

class open (db3.open):
	def __init__ (self, dbname, user, password, host = '127.0.0.1', port = 5432):
		self.conn = psycopg2.connect (host=host, dbname=dbname, user=user, password=password, port = port)
		self.create_cursor ()
	
	def field_names (self):
		return [x.name for x in self.description]



import xmlrpc.client
from . import timeoutsocket

timeoutsocket.setDefaultSocketTimeout(3)

servers = []

class ServerNotDefined (Exception): pass

def setup (serverlist, timeout = 3):
	global servers
	servers = serverlist
	timeoutsocket.setDefaultSocketTimeout(timeout)

def call (method, args):
	global servers
	if not servers:
		raise ServerNotDefined

	for server in servers:
		try:
			s = xmlrpc.client.Server ("http://" + server)
			resp = getattr (s, method) (*args)
			return resp
		except:
			pass
	raise



import asyncore
import socket
from .select_trigger import trigger

the_trigger = None
the_logger = None

def start_trigger (logger = None):
	global the_trigger, the_logger
	the_logger = logger
	if the_trigger is None:
		the_trigger = trigger (logger)

def wakeup (thunk = None):
	global the_trigger, the_logger
		
	if the_trigger is None:
		if thunk:
			try:
				thunk ()
			except:
				if the_logger:
					the_logger.trace ('wakeup')
		return		

	try:
		the_trigger.pull_trigger(thunk)
	except OSError as why:
		if why.errno == 32:
			the_trigger.close ()
			the_trigger = trigger ()
			the_trigger.pull_trigger(thunk)
	except socket.error:
		the_trigger.close ()
		the_trigger = trigger ()
		the_trigger.pull_trigger(thunk)
			
def wakeselect ():
	for fd, obj in list(asyncore.socket_map.items()):
		if hasattr(obj, "pull_trigger"):
			obj.pull_trigger()



import sys

PY_MAJOR_VERSION = sys.version_info [0]

if PY_MAJOR_VERSION >= 3:
	def makes (arg, encoding = None):
		argtype = type (arg)
		if argtype is str:
			return arg.strip ()
		if argtype is bytes and encoding:
			return arg.decode (encoding)
		if arg is None: 
			return ""		
		if argtype is int:
			return str (arg)
			
		raise TypeError ("Must be String, Int, None or Bytes with encoding")	
				
else:		
	def makes (arg, encoding = "utf8"):
		argtype = type (arg)
		if argtype is unicode:
			return arg.encode ("utf8").strip ()
		if argtype is str:
			if encoding in (None, "utf8"):
				return arg.strip ()
			return arg.decode (encoding).encode ("utf8")
		if arg is None:
			return ""
		if argtype is int:
			return str (arg)		
			
		raise TypeError ("Must be Unicode, Int, None or String with encoding")	
		



import time
try:
	from h2.exceptions import FlowControlError
except ImportError:
	pass	
from aquests.lib import producers
import asynchat

class h2header_producer:
	def __init__ (self, stream_id, headers, producer, encoder, lock):
		# DO NOT set self.stream_id
		# if set, http2_producer_fifo try to re-sort and will raise error
		self.__stream_id = stream_id
		with lock:
			encoder.send_headers (
				stream_id = self.__stream_id,
				headers = headers,
				end_stream = producer is None
			)
			self.data_to_send = encoder.data_to_send ()
	
	def get_size (self):
		return 0
		
	def __repr__ (self):
		return "<h2header_producer stream_id:%d>" % (self.__stream_id,)
	
	def more (self):		
		data_to_send, self.data_to_send = self.data_to_send, b''		
		return data_to_send


class h2frame_producer:
	SIZE_BUFFER = 16384
	MIN_RFCW = 4096
	MIN_IBFCW = 65535
	def __init__ (self, stream_id, depends_on, weight, producer, encoder, lock, trailers = False):
		self.stream_id = stream_id
		self.depends_on = depends_on
		self.weight = weight
		self.producer = producer # globbing_producer
		self.encoder = encoder
		self._lock = lock	
		self.trailers = trailers	
		
		self._buf = b""
		self._end_stream = False		
		self._last_sent = time.time ()
		self._ready = None
		self._frame = b''
		
		if hasattr (producer, "ready"):
			self._ready = producer.ready
			producer.ready = None			
	
	def reset_stream (self,error_code = 3):
		self.encoder.reset_stream (self.stream_id, error_code = 3)
		self._frame = self.encoder.data_to_send ()
		self._buf = b''
		self._end_stream = True		 
	
	def exhausted (self):
		return self._end_stream and not self._buf and not self._frame
		
	def ready (self):		
		if self.exhausted ():			
			return True
			
		if self._frame:
			return True
		
		lfcw = self.encoder.local_flow_control_window (self.stream_id)
		#print (">>>LFCW", self.stream_id, lfcw)		
		if lfcw == 0:
			# flow control error, graceful close
			if time.time () - self._last_sent > 10:
				self.reset_stream (3)				 
				return True
			return False
				
		avail_data_length = min (self.SIZE_BUFFER, lfcw)
		if not self._buf:
			if self._ready and not self._ready ():
				return False
			self._buf = self.producer.more ()
			self._end_stream = self.producer.exhausted ()
			#assert self._end_stream == self.is_end_stream (self._buf)
		data, self._buf = self._buf [:avail_data_length], self._buf [avail_data_length:]
				
		# try build frame
		is_end_stream = self._end_stream and not self._buf
		with self._lock:		
			self.encoder.send_data (
				stream_id = self.stream_id,
				data = data,
				end_stream = is_end_stream and not self.trailers
			)
			if is_end_stream and self.trailers:
				self.encoder.send_headers (
					stream_id = self.stream_id,
					headers = self.trailers,
					end_stream = True
				)			
			self._frame = self.encoder.data_to_send ()
			
		return True
	
	def get_size (self):			
		return self._ready and -1 or self.producer.get_size ()
		
	def __repr__ (self):
		return "<h2frame_producer stream_id:%d, weight:%d, depends_on:%d>" % (self.stream_id, self.weight, self.depends_on)
					 	
	def more (self):
		#print ('++++++++++++++more', len (self._frame), self.exhausted ())
		self._last_sent = time.time ()
		if self.exhausted ():
			return b''
		frame, self._frame = self._frame, b''		
		return frame
	